@@ -3,6 +3,24 @@ require 'rack/utils'
 require 'rack/mock'
 
 describe Rack::Utils do
+  def kcodeu
+    one8 = RUBY_VERSION.to_f < 1.9
+    default_kcode, $KCODE = $KCODE, 'U' if one8
+    yield
+  ensure
+    $KCODE = default_kcode if one8
+  end
+
+  should "round trip binary data" do
+    r = [218, 0].pack 'CC'
+    if defined?(::Encoding)
+      z = Rack::Utils.unescape(Rack::Utils.escape(r), Encoding::BINARY)
+    else
+      z = Rack::Utils.unescape(Rack::Utils.escape(r))
+    end
+    r.should.equal z
+  end
+
   should "escape correctly" do
     Rack::Utils.escape("fo<o>bar").should.equal "fo%3Co%3Ebar"
     Rack::Utils.escape("a space").should.equal "a+space"
@@ -21,23 +39,27 @@ describe Rack::Utils do
 
   if RUBY_VERSION[/^\d+\.\d+/] == '1.8'
     should "escape correctly for multibyte characters if $KCODE is set to 'U'" do
-      default_kcode, $KCODE = $KCODE, 'U'
-
-      matz_name = "\xE3\x81\xBE\xE3\x81\xA4\xE3\x82\x82\xE3\x81\xA8".unpack("a*")[0] # Matsumoto
-      matz_name.force_encoding("UTF-8") if matz_name.respond_to? :force_encoding
-      Rack::Utils.escape(matz_name).should.equal '%E3%81%BE%E3%81%A4%E3%82%82%E3%81%A8'
-      matz_name_sep = "\xE3\x81\xBE\xE3\x81\xA4 \xE3\x82\x82\xE3\x81\xA8".unpack("a*")[0] # Matsu moto
-      matz_name_sep.force_encoding("UTF-8") if matz_name_sep.respond_to? :force_encoding
-      Rack::Utils.escape(matz_name_sep).should.equal '%E3%81%BE%E3%81%A4+%E3%82%82%E3%81%A8'
-
-      $KCODE = default_kcode
+      kcodeu do
+        matz_name = "\xE3\x81\xBE\xE3\x81\xA4\xE3\x82\x82\xE3\x81\xA8".unpack("a*")[0] # Matsumoto
+        matz_name.force_encoding("UTF-8") if matz_name.respond_to? :force_encoding
+        Rack::Utils.escape(matz_name).should.equal '%E3%81%BE%E3%81%A4%E3%82%82%E3%81%A8'
+        matz_name_sep = "\xE3\x81\xBE\xE3\x81\xA4 \xE3\x82\x82\xE3\x81\xA8".unpack("a*")[0] # Matsu moto
+        matz_name_sep.force_encoding("UTF-8") if matz_name_sep.respond_to? :force_encoding
+        Rack::Utils.escape(matz_name_sep).should.equal '%E3%81%BE%E3%81%A4+%E3%82%82%E3%81%A8'
+      end
     end
 
     should "unescape multibyte characters correctly if $KCODE is set to 'U'" do
-      default_kcode, $KCODE = $KCODE, 'U'
-      Rack::Utils.unescape('%E3%81%BE%E3%81%A4+%E3%82%82%E3%81%A8').should.equal(
-        "\xE3\x81\xBE\xE3\x81\xA4 \xE3\x82\x82\xE3\x81\xA8".unpack("a*")[0])
-      $KCODE = default_kcode
+      kcodeu do
+        Rack::Utils.unescape('%E3%81%BE%E3%81%A4+%E3%82%82%E3%81%A8').should.equal(
+          "\xE3\x81\xBE\xE3\x81\xA4 \xE3\x82\x82\xE3\x81\xA8".unpack("a*")[0])
+      end
+    end
+  end
+
+  should "escape objects that responds to to_s" do
+    kcodeu do
+      Rack::Utils.escape(:id).should.equal "id"
     end
   end
 
@@ -46,6 +68,16 @@ describe Rack::Utils do
       Rack::Utils.escape("ø".encode("ISO-8859-1")).should.equal "%F8"
     end
   end
+  
+  should "not hang on escaping long strings that end in % (http://redmine.ruby-lang.org/issues/5149)" do
+    lambda {
+      timeout(1) do
+        lambda {
+          URI.decode_www_form_component "A string that causes catastrophic backtracking as it gets longer %"
+        }.should.raise(ArgumentError)
+      end
+    }.should.not.raise(Timeout::Error)
+  end
 
   should "escape path spaces with %20" do
     Rack::Utils.escape_path("foo bar").should.equal  "foo%20bar"
@@ -185,22 +217,22 @@ describe Rack::Utils do
     # unordered hash. Test that build_nested_query performs the inverse
     # function of parse_nested_query.
     [{"foo" => nil, "bar" => ""},
-     {"foo" => "bar", "baz" => ""},
-     {"foo" => ["1", "2"]},
-     {"foo" => "bar", "baz" => ["1", "2", "3"]},
-     {"foo" => ["bar"], "baz" => ["1", "2", "3"]},
-     {"foo" => ["1", "2"]},
-     {"foo" => "bar", "baz" => ["1", "2", "3"]},
-     {"x" => {"y" => {"z" => "1"}}},
-     {"x" => {"y" => {"z" => ["1"]}}},
-     {"x" => {"y" => {"z" => ["1", "2"]}}},
-     {"x" => {"y" => [{"z" => "1"}]}},
-     {"x" => {"y" => [{"z" => ["1"]}]}},
-     {"x" => {"y" => [{"z" => "1", "w" => "2"}]}},
-     {"x" => {"y" => [{"v" => {"w" => "1"}}]}},
-     {"x" => {"y" => [{"z" => "1", "v" => {"w" => "2"}}]}},
-     {"x" => {"y" => [{"z" => "1"}, {"z" => "2"}]}},
-     {"x" => {"y" => [{"z" => "1", "w" => "a"}, {"z" => "2", "w" => "3"}]}}
+      {"foo" => "bar", "baz" => ""},
+      {"foo" => ["1", "2"]},
+      {"foo" => "bar", "baz" => ["1", "2", "3"]},
+      {"foo" => ["bar"], "baz" => ["1", "2", "3"]},
+      {"foo" => ["1", "2"]},
+      {"foo" => "bar", "baz" => ["1", "2", "3"]},
+      {"x" => {"y" => {"z" => "1"}}},
+      {"x" => {"y" => {"z" => ["1"]}}},
+      {"x" => {"y" => {"z" => ["1", "2"]}}},
+      {"x" => {"y" => [{"z" => "1"}]}},
+      {"x" => {"y" => [{"z" => ["1"]}]}},
+      {"x" => {"y" => [{"z" => "1", "w" => "2"}]}},
+      {"x" => {"y" => [{"v" => {"w" => "1"}}]}},
+      {"x" => {"y" => [{"z" => "1", "v" => {"w" => "2"}}]}},
+      {"x" => {"y" => [{"z" => "1"}, {"z" => "2"}]}},
+      {"x" => {"y" => [{"z" => "1", "w" => "a"}, {"z" => "2", "w" => "3"}]}}
     ].each { |params|
       qs = Rack::Utils.build_nested_query(params)
       Rack::Utils.parse_nested_query(qs).should.equal params
@@ -211,7 +243,18 @@ describe Rack::Utils do
       message.should.equal "value must be a Hash"
   end
 
-  should "should escape html entities [&><'\"/]" do
+  should "parse query strings that have a non-existent value" do
+    key = "post/2011/08/27/Deux-%22rat%C3%A9s%22-de-l-Universit"
+    Rack::Utils.parse_query(key).should.equal Rack::Utils.unescape(key) => nil
+  end
+
+  should "build query strings without = with non-existent values" do
+    key = "post/2011/08/27/Deux-%22rat%C3%A9s%22-de-l-Universit"
+    key = Rack::Utils.unescape(key)
+    Rack::Utils.build_query(key => nil).should.equal Rack::Utils.escape(key)
+  end
+
+  should "escape html entities [&><'\"/]" do
     Rack::Utils.escape_html("foo").should.equal "foo"
     Rack::Utils.escape_html("f&o").should.equal "f&amp;o"
     Rack::Utils.escape_html("f<o").should.equal "f&lt;o"
@@ -222,6 +265,27 @@ describe Rack::Utils do
     Rack::Utils.escape_html("<foo></foo>").should.equal "&lt;foo&gt;&lt;&#x2F;foo&gt;"
   end
 
+  should "escape html entities even on MRI when it's bugged" do
+    test_escape = lambda do
+      kcodeu do
+        Rack::Utils.escape_html("\300<").should.equal "\300&lt;"
+      end
+    end
+
+    if RUBY_VERSION.to_f < 1.9
+      test_escape.call
+    else
+      test_escape.should.raise(ArgumentError)
+    end
+  end
+
+  if "".respond_to?(:encode)
+    should "escape html entities in unicode strings" do
+      # the following will cause warnings if the regex is poorly encoded:
+      Rack::Utils.escape_html("☃").should.equal "☃"
+    end
+  end
+
   should "figure out which encodings are acceptable" do
     helper = lambda do |a, b|
       Rack::Request.new(Rack::MockRequest.env_for("", "HTTP_ACCEPT_ENCODING" => a))
